use tokio::net::{TcpListener, TcpStream};
use tokio_tungstenite::tungstenite::protocol::WebSocketConfig;
use tokio_tungstenite::tungstenite::protocol::Message;
use tokio_tungstenite::WebSocketStream;
use futures_util::{StreamExt, SinkExt};
use futures_util::stream::SplitSink;
use serde::Serialize;

#[derive(Debug, Serialize)]
pub enum Hemisphere {
    North,
    South,
    East,
    West,
}

#[derive(Debug, Serialize)]
pub struct Latitude {
    degrees: u8,
    minutes: u8,
    seconds: f32,
    hemisphere: Hemisphere,
}

#[derive(Debug, Serialize)]
pub struct Longitude {
    degrees: u8,
    minutes: u8,
    seconds: f32,
    hemisphere: Hemisphere,
}

pub struct Solution {
    pub latitude: Latitude,
    pub longitude: Longitude,
    pub altitude: Option<f32>,
    pub speed: Option<f32>,
    pub direction: Option<f32>,
}

struct DeviceLocation {
    device_id: String,
    solution: Solution,
}

async fn stream_locations(mut writer: SplitSink<WebSocketStream<TcpStream>, Message>) -> anyhow::Result<()> {
    let device_id = "TESTTEST6355432da4242TESTTEST";
    let mut lat = 44.76027;
    let mut long = 20.47723;

    for i in 1..100 {
        lat = lat + if i > 50 { -0.001 } else { 0.001 };
        long = long + if i > 50 { -0.001 } else { 0.001 };
        let msg = Message::Text(
            format!("{{\"device_id\":\"{}\", \"lat\":{}, \"long\":{}}}",
                    device_id, lat, long,
            )
        );
        let _ = writer.send(msg).await?;
        tokio::time::sleep(tokio::time::Duration::from_secs(5)).await;
    }

    Ok(())
}

#[tokio::main]
async fn main() -> anyhow::Result<()> {
    let addr = "127.0.0.1:19999";

    let tcp_server = TcpListener::bind(&addr).await?;

    println!("listening on {}", addr);

    while let Ok((tcp_stream, addr)) = tcp_server.accept().await {
        println!("Got a connection from: {}", addr);

        let wsconfig = WebSocketConfig {
            max_send_queue: Some(10),
            max_message_size: None,
            max_frame_size: None,
            accept_unmasked_frames: true,
        };

        let ws_stream = tokio_tungstenite::accept_async_with_config(tcp_stream, Some(wsconfig))
            .await
            .expect("Error during the websocket handshake occurred");

        let (writer, mut reader) = ws_stream.split();

        let token_optres = reader.next().await;

        if let Some(Ok(token)) = token_optres {
            let local_token = include_str!("../secret/token").trim();
            println!("{}", local_token);
            if token.into_text()? == local_token {
                println!("streaming locations");
                match stream_locations(writer).await {
                    Ok(_) => println!("closing connection to client ..."),
                    Err(error) => println!("connection to client raised an error: ({})", error),
                }
            }
            else {
                println!("Closing client connection due to bad token!");
                let mut s = writer.reunite(reader)?;
                s.close(None).await?;
            }
        }

    }

    Ok(())
}
